@@ -1,6 +1,6 @@
 # coding: utf-8
 from __future__ import unicode_literals
-from devpi_common.metadata import get_pyversion_filetype, splitbasename
+from devpi_common.metadata import get_pyversion_filetype
 from devpi_common.metadata import get_sorted_versions
 from devpi_common.viewhelp import iter_toxresults
 from devpi_server.log import threadlog as log
@@ -31,14 +31,6 @@ class ContextWrapper(object):
     def __getattr__(self, name):
         return getattr(self.context, name)
 
-    @reify
-    def stage(self):
-        stage = self.model.getstage(self.username, self.index)
-        if not stage:
-            raise HTTPNotFound(
-                "The stage %s/%s could not be found." % (self.username, self.index))
-        return stage
-
     @reify
     def versions(self):
         versions = self.stage.list_versions(self.name)
@@ -46,15 +38,6 @@ class ContextWrapper(object):
             raise HTTPNotFound("The project %s does not exist." % self.name)
         return get_sorted_versions(versions)
 
-    @reify
-    def verdata(self):
-        verdata = self.stage.get_versiondata(self.name, self.version)
-        if not verdata and self.versions:
-            raise HTTPNotFound(
-                "The version %s of project %s does not exist." % (
-                    self.version, self.name))
-        return verdata
-
     @reify
     def linkstore(self):
         return self.stage.get_linkstore_perstage(self.name, self.version)
@@ -64,7 +47,7 @@ def get_doc_path_info(context, request):
     relpath = request.matchdict['relpath']
     if not relpath:
         raise HTTPFound(location="index.html")
-    name = context.name
+    name = context.projectname
     version = context.version
     if version != 'latest':
         versions = [version]
@@ -98,7 +81,7 @@ def doc_show(context, request):
     """ Shows the documentation wrapped in an iframe """
     context = ContextWrapper(context)
     stage = context.stage
-    name, version = context.name, context.version
+    name, version = context.projectname, context.version
     doc_path, relpath = get_doc_path_info(context, request)
     return dict(
         title="%s-%s Documentation" % (name, version),
@@ -130,7 +113,7 @@ def notfound(request):
                 qs = request.query_string
                 if qs:
                     qs = '?' + qs
-                return HTTPFound(location=nonslashpath + qs)
+                return HTTPFound(location=request.application_url + nonslashpath + qs)
     request.response.status = 404
     return dict(msg=request.exception)
 
@@ -181,7 +164,8 @@ def make_history_view_item(request, log_item):
     result = {}
     result['what'] = _what_map.get(log_item['what'], log_item['what'])
     result['who'] = log_item['who']
-    result['when'] = format_timetuple(log_item['when'])
+    if log_item['what'] != 'overwrite':
+        result['when'] = format_timetuple(log_item['when'])
     for key in ('dst', 'src'):
         if key in log_item:
             result[key] = dict(
@@ -247,6 +231,7 @@ def get_toxresults_info(linkstore, for_link, newest=True):
     for toxlink, toxenvs in iter_toxresults(toxlinks, load_toxresult, newest=newest):
         if toxenvs is None:
             log.error("Couldn't parse test results %s." % toxlink)
+            continue
         for toxenv in toxenvs:
             info = dict(
                 basename=toxlink.basename,
@@ -380,17 +365,17 @@ def index_get(context, request):
 def project_get(context, request):
     context = ContextWrapper(context)
     try:
-        releaselinks = context.stage.get_releaselinks(context.name)
+        releaselinks = context.stage.get_releaselinks(context.projectname)
     except context.stage.UpstreamError as e:
         log.error(e.msg)
         raise HTTPBadGateway(e.msg)
-    if not releaselinks:
-        raise HTTPNotFound("The project %s does not exist." % context.name)
     versions = []
     seen = set()
     for release in releaselinks:
         user, index = release.entrypath.split("/", 2)[:2]
-        name, version = splitbasename(release)[:2]
+        name, version = release.projectname, release.version
+        if version == 'XXX':
+            continue
         seen_key = (user, index, name, version)
         if seen_key in seen:
             continue
@@ -412,12 +397,13 @@ def project_get(context, request):
     accept="text/html", request_method="GET",
     renderer="templates/version.pt")
 def version_get(context, request):
+    """ Show version for the precise stage, ignores inheritance. """
     context = ContextWrapper(context)
     user, index = context.username, context.index
     name, version = context.name, context.version
     stage = context.stage
     try:
-        verdata = context.verdata
+        verdata = context.get_versiondata(perstage=True)
     except stage.UpstreamError as e:
         log.error(e.msg)
         raise HTTPBadGateway(e.msg)
@@ -455,6 +441,10 @@ def version_get(context, request):
         url=request.route_url(
             "/{user}/{index}/+simple/{name}",
             user=context.username, index=context.index, name=context.name)))
+    if stage.has_pypi_base(name):
+        nav_links.append(dict(
+            title="PyPI page",
+            url="https://pypi.python.org/pypi/%s" % name))
     return dict(
         title="%s/: %s-%s metadata and description" % (stage.name, name, version),
         content=get_description(stage, name, version),
